using System;
using System.Threading;

namespace IOWarrior
{
	/// <summary>
	/// Main Class for accessing (read/write) from or to the IOWarrior
	/// </summary>
	public class IOWKit
	{
		protected int iowHandle = 0; 
		
		private int nColumns = 0;
		private int nRows = 0;
		private bool bShowCursor = false;
		private bool bCursorBlinking = false;
		private bool bIREnabled = false;
		
		// public read-only Properties
		public bool IREnabled
		{
			get{ return bIREnabled; }
		}
		
		public bool LCDEnabled
		{
			get{ return this.iowHandle == 0 ? false : true; }
		}

		public int NrOfRows
		{
			get{ return this.nRows; }
		}

		public int NrOfColumns
		{
			get{ return this.nColumns; }
		}

		public string ProductId
		{
			get	
			{ 
				short sProductId = IOWLib.IowKitGetProductId( this.iowHandle ); 
				return sProductId.ToString("x"); // hexadezimal, 1500 = IOWarrior40; 1501 = IOWarrior24 
			}
		}

		public string SerialNumber
		{
			get 
			{
				short[] sn = new short[8];
				if(IOWLib.IowKitGetSerialNumber( this.iowHandle, ref sn[0] ))
				{
					string serial = string.Empty;
					foreach(char c in sn)
						serial += c;
					return serial;
				}
				else
					return "unknown";
			}
		}

		public enum AlignType { Left, Center, Right };

		// IR Stuff
		public delegate void IREventHandler(object source, IREventArgs args);
		public event IREventHandler OnIREvent;
		private Thread IRThread;

		/// <summary>
		/// Enables the IR Function
		/// </summary>
		/// <returns></returns>
		public bool EnableIR()
		{
			if(this.LCDEnabled == false)
				return false;

			// Check IOWarrior Version - IR is only available on IOWarrior24
			if(this.ProductId != "1501") // 1501 = IOWarrior24
				throw new Exception("IR functionality is only available for IOWarrior24");

			// To enable IR, simply send a Report with ReportID 0xC and the "enabled" flag (see 5.10.5 on IOWarrior Datasheet)
			int ret = SendReportId(0xc, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0);

			// We must use a thread for reading IR Data here, otherwise the application will wait 
			// each time we want to read from the IR Port
			// have a look at the event "OnIREvent" here and the "MyOnIREvent" at the Example Application (Form1.cs)
			IRThread = new Thread(new ThreadStart(ReadIrKeys));
			IRThread.Name = "IRReadingThread";
			IRThread.Start();
			bIREnabled = true;
			return ret > 0 ? true : false;
		}

		/// <summary>
		/// Disables the IR Function
		/// </summary>
		/// <returns>true, if IR was disabled, otherwise false</returns>
		public bool DisableIR()
		{
			if( bIREnabled == false )
				return false;

			bIREnabled = false;
			IOWLib.IowKitCancelIo( this.iowHandle, 1);
			int ret = SendReportId(0xc, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0);
			IRThread.Abort();

			return ret > 0 ? true : false;
		}

		/// <summary>
		/// This function is called by a thread and should not be called directly
		/// </summary>
		private void ReadIrKeys()
		{
			byte key = 0; 
			byte rc = 0;
			byte toggle = 0;

			while( bIREnabled )
			{
				// we must set the timeout here, specially for the IR read function.
				// otherwise the application could never exit because the IRReadThread could still be running

				byte[] byteInp = new byte[8];
				IOWLib.IowKitRead(iowHandle, 1, ref byteInp[0], 8);
				key = byteInp[1];
				rc = (byte)(byteInp[2] & 0x1f); // 0x1f = binary 00011111 - these are the "address" bits
				toggle = (byte)(byteInp[2] & 0x20); // 0x20 = binary 00100000 - get the toggle bit

				IREventArgs e = new IREventArgs( key, rc, toggle );

				// before sending the event, check if IR Function is still enabled 
				// also check if we REALLY received a key, otherwise it was the iow-timeout we set at Init()
				if( bIREnabled && (byteInp[2] != 0x0))
					OnIREvent( this, e );
	
			}
		}

		/// <summary>
		/// Initializes the LCD Display with standard values (no cursor, no blinking and without IR)
		/// </summary>
		/// <param name="columns">Columns of LCD</param>
		/// <param name="rows">Rows of LCD</param>
		/// <returns>true, if successfull, otherwise false</returns>
		public bool Init(int columns, int rows)
		{
			return Init(columns, rows, false, false);
		}

		/// <summary>
		/// Initializes the LCD Display
		/// </summary>
		/// <param name="columns">Columns of LCD</param>
		/// <param name="rows">Rows of LCD</param>
		/// <param name="showCursor">Wheter or not the cursor (_) should be displayed</param>
		/// <param name="cursorBlinking">Wheter or not the a blinking cursor should be displayed</param>
		/// <returns>true, if successfull, otherwise false</returns>
		public bool Init(int columns, int rows, bool showCursor, bool cursorBlinking)
		{
			// check if display is open
			if( iowHandle != 0) 
				throw new Exception("you cannot initialize your display twice, call the Disable function first");

			// opens device and returns the handle
			iowHandle = IOWLib.IowKitOpenDevice();
			
			if( iowHandle == 0 )
				return false;

			// set member variables
			nColumns = columns;
			nRows = rows;
			bShowCursor = showCursor;
			bCursorBlinking = cursorBlinking;

			// enables LCD Special mode funtion, see 5.10.2 on IOW-Datasheet
			int ret = SendReportId(0x4, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0);
			if (ret == 0)
				return false;

			// new ReportID to finally initialize the LCD
			// byte1: data length to be send (3 bytes in our case)
			// byte2: clear Display if value = 0x1
			// byte3: Display on / off, Cursor and Cursor blinking
			// byte4: function set (interface datalength, display lines and character font)
			// the order of the data instruction bytes (byte 2-7) may vary, you can also 
			// add other data instruction bytes if you want (you must increase the data length then)
			// for more information see commandset description of the HD44780

			byte functionSetVal = GenerateInitialFunctionSetValue();
			byte modeVal = GenerateInitialModeValue();

			ret = SendReportId( 0x5, 0x3, 0x1, modeVal, functionSetVal, 0x0, 0x0, 0x0);
			
			return ret > 0 ? true : false;
		}

		/// <summary>
		/// Disables LCD and IR Special Mode Function and closes the device
		/// </summary>
		/// <returns></returns>
		public void Quit()
		{
			// retrn if LCD is still disabled
			if( iowHandle == 0)
				return;

			// Disables IR
			DisableIR();

			// disables LCD Special mode function (see 5.10.2)
			SendReportId(0x4, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0);

			// finally close device
			IOWLib.IowKitCloseDevice( iowHandle );
			iowHandle = 0;
		}

		/// <summary>
		/// Sets the cursor position on the display
		/// Indexes will start at 1, not 0!
		/// </summary>
		/// <param name="x">X-Coordinate (column)</param>
		/// <param name="y">Y-Coordinate (row)</param>
		/// <returns>true if position could be set, otherwise false</returns>
		public bool SetPosition(int x, int y)
		{
			byte addr = 0;

			if(( x > nColumns ) || (y > nRows))
				return false; // impossible position

			// if nRows is 1, it's very simple
			if( nRows == 1)
				addr = (byte)x;
			// otherwise we have to calculate our address
			// remember: the HD44780 only knows 2 lines; at 4-lines-displayes, 
			//		line 1+3 and 2+4 will be handled as one single line
			// further notice: the second line for the HD44780 ALWAYS is at 0x40,
			//		no matter how many columns each row has (as seen on the datasheet for my LCD)
			else
			{
				if( y == 1 )
					addr = 0;
				if (y == 2 )
					addr = 0x40;
				if (y == 3 )
					addr = (byte)nColumns;
				if (y == 4 )
					addr = (byte)(0x40 + nColumns);

				addr += (byte)(x - 1);
			}

			// to let the controller know, that we want to set the DD RAM Address, we have to 
			// set the "DD RAM Address Set" Instruction Bit: 01000000 = 0x80 
			addr = (byte)((int)addr | 0x80);

			// finally send the ReportId. We only have to send one Instruction,
			// so the second Parameter (d1) is 0x1 
			// see the IOWarrior Datasheet, Chapter 5.10.2:
			//	RS(Register Selectbit) = 0 - we want to write in the Instruction Register
			int ret = SendReportId(0x5, 0x1, addr, 0x0, 0x0, 0x0, 0x0, 0x0);

			return ret > 0 ? true : false;
		}

		/// <summary>
		/// Writes one character to the given position
		/// </summary>
		/// <param name="x">X-Coordinate (column)</param>
		/// <param name="y">Y-Coordinate (row)</param>
		/// <param name="character">the character to be written</param>
		/// <returns>true if successfull, otherwise false</returns>
		public bool WriteAt(int x, int y, char character)
		{
			// first, set the position
			bool bRet = SetPosition(x, y);
			if( bRet == false ) // something wrent wong
				return false;

			// now write the character. This time, we do not write to the instruction 
			// register, so the flag byte has to be binary 10000000 ( = 0x80)
			// while we still have to say, how many bytes we will write.

			// we only want to write one character (one byte) so we can simply
			// do:
			byte flagByte = 0x80 | 0x1;

			// because the charset on the LCD does not match the ASCII codes, 
			// we have to translate some characters with the GetLCDByteValue Function
			byte characterByte = GetLCDByteValue( character );
			int ret = SendReportId(0x5, flagByte, characterByte, 0x0, 0x0, 0x0, 0x0, 0x0);

			return ret > 0 ? true : false;
		}

		/// <summary>
		/// Writes a text to the given Display line number. 
		/// If the text is greater than the number of columns, the text will be cutted.
		/// You can use the WriteScroll or WriteMarquee Functions instead.
		/// </summary>
		/// <param name="lineNumber"></param>
		/// <param name="text"></param>
		/// <returns>true, if successfull, otherwise false</returns>
		public bool WriteLine(int lineNumber, string text)
		{
			return WriteLine( lineNumber, text, AlignType.Left );
		}

		/// <summary>
		/// Writes a text to the given Display line number. 
		/// If the text is greater than the number of columns, the text will be cutted.
		/// You can use the WriteScroll or WriteMarquee Functions instead.
		/// </summary>
		/// <param name="lineNumber"></param>
		/// <param name="text"></param>
		/// <param name="align">Enum for the text-align, one of Left, Right or Center</param>
		/// <returns>true, if successfull, otherwise false</returns>
		public bool WriteLine(int lineNumber, string text, AlignType align)
		{
			// check for Special Chars placeholders
			// (use regex here if you don't like this kind of replacing)
			for(int i = 0; i < 8; i++)
				text = text.Replace("{SP" + i + "}", ((char)i).ToString());

			// cut the text if it is too long..
			if( text.Length > nColumns)
				text = text.Substring(0, nColumns);

			int pos = 1;
			if( align == AlignType.Right)
				pos = nColumns - text.Length + 1;
			if( align == AlignType.Center)
				pos = (nColumns - text.Length) / 2 + 1;

			// this code replaces the ClearLine() function.
			// If we would this function, the LCD line may flicker.
			// 
			String strPre = string.Empty;
			String strPost = string.Empty;
			for(int i = 1; i < pos; i++)
				strPre += " ";
			for(int i = 20; i > ((string)(strPre + text)).Length; i--)
				strPost += " ";

			// first, set the position
			SetPosition(1, lineNumber);

			// now, write the string out
			return WriteString( strPre + text + strPost );
		}

		/// <summary>
		/// This function is usually called by other Functions like WriteLine etc. but you 
		/// can still use it when you set the Cursor Position with the SetPosition function.
		/// Warning: This function does not take care about the HD44780 line arrangement!
		/// </summary>
		/// <param name="text"></param>
		/// <returns></returns>
		public bool WriteString( string text )
		{
			// the HD44780 can only write 6 bytes at once, so we have to split the string
			// (we also could write out each single byte, but this would be a bit too slow)

			int lastPos = 0;
			string currentString = string.Empty;
			do
			{
				currentString = text.Substring(lastPos, (lastPos + 6) < text.Length ? 6 : (text.Length - lastPos));
				lastPos += 6;

				char[] charsOut = new char[6];						// upperbound is always 6
				char[] usedChars = currentString.ToCharArray();		// upperbound unfortunatelly may vary

				for(int i = 0; i < usedChars.Length; i++)
					charsOut[i] = usedChars[i];

				// like the WriteAt function, we have to write to the DataRegister
				byte flagByte = (byte)(0x80 | currentString.Length); // 0x80 = binary 10000000; indicates that we want to write to the data register
				int ret = SendReportId(0x5, flagByte, GetLCDByteValue( charsOut[0] ), GetLCDByteValue( charsOut[1] ), GetLCDByteValue( charsOut[2] ), GetLCDByteValue( charsOut[3] ), GetLCDByteValue( charsOut[4] ), GetLCDByteValue( charsOut[5] ));
				if( ret == 0)
					return false;

			} 
			while((text.Length - lastPos) > 0 );

			return true;
		}

		/// <summary>
		/// Clears the given Line
		/// </summary>
		/// <param name="lineNumber"></param>
		/// <returns></returns>
		public bool ClearLine( int lineNumber )
		{
			string str = new string(' ', nColumns);

			// we do NOT use the WriteLine function here, because doing this
			// would end up in a stack overflow because of recursion of ClearLine
			SetPosition(1, lineNumber);
			bool ret = WriteString(str);
			SetPosition(1, lineNumber);
			return ret;
		}

		/// <summary>
		/// Clears the whole LCD
		/// </summary>
		/// <returns>true if successfull, otherwise false</returns>
		public bool ClearScreen()
		{
			// this function is pretty simple.
			// we have to send a ReportId 5 into the Instruction Register with a single byte.
			// because Instruction Register = 00000000 and the nr of bytes we will send is 00000001
			// the flagByte after logically OR'ing it is: 00000001 = 0x1 
			byte flagByte = 0x1;

			// the value for the instruction is 00000001 (as seen on LCD Datasheet), so the second byte-Value is 0x1 too
			int ret = SendReportId(0x5, flagByte, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0);

			return ret > 0 ? true : false;
		}

		/// <summary>
		/// Sets the cursor mode for the LCD.
		/// </summary>
		/// <param name="bShowCursor">Indicates if the underline-cursor should be displayed</param>
		/// <param name="bCursorBlinking">Indicates if the blinking-cursor should be displayed</param>
		/// <returns>true if successfull, otherwise: false</returns>
		public bool SetCursorMode( bool bShowCursor, bool bCursorBlinking)
		{
			byte b = 0xc; // 0xc = binary 00001000 & 00000100: 00001000 is Instruction for setting Display/Cursor Options and 00000100 the value for Display On (should be always on)

			if(bShowCursor)
				b = (byte)(b | 0x2); // 0x3 = binary 00000010: Cursor On/Off
			if(bCursorBlinking)
				b = (byte)(b | 0x1); // 0x1 = binary 00000001: Cursor blinking yes/no

			int ret = SendReportId(0x5, 0x1, b, 0x0, 0x0, 0x0, 0x0, 0x0);
			return ret > 0 ? true : false;
		}

		/// <summary>
		/// uploads a special character for ascii code 0 to 7. Use {SP0} to {SP7} in
		///  WriteLine() to write this character on your display
		/// </summary>
		/// <param name="index">The ascii code = character Number for the special character. Can only be 0 to 7</param>
		/// <param name="dotMatrix">The boolean character representation of the character - has to be a bool[5][8] field</param>
		/// <returns>true if successfull, otherwise: false</returns>
 		public bool UploadSpecialChar( int index, bool[][] dotMatrix)
		{
			if( (index < 0) || (index > 7))
				return false;

			// byteToWrite includes linenumber and character data for this line
			// so firstly we set the line number (bit 7,6,5) - we have to add 0x8 (bin 0001000) for each index
			// to tell the HD44780 that we want to set the CG RAM Address, we have to send this Instruction: 0x40 ( = 0100000)
			// look at your LCD / HD44780 datasheet for more information.
			byte address = (byte)((byte)(index * 0x8) | 0x40);

			// we have to write each character line seperately
			for(int j = 0; j < 8; j++)
			{
				byte line = 0x0;
				for(int i = 0; i < 5; i++)
				{
					if( dotMatrix[i][j] == true) // if bit is set
						line = (byte)(line | (byte)Math.Pow(2, 5-i-1));
				}

				// first we have to set the CC RAM Address. 
				// We have to increment the address for each character line
				// address = character number (index) and the instruction bit for "CG RAM Address Set" (0x40 = 01000000))
				SendReportId(0x5, 0x1, address++, 0x0, 0x0, 0x0, 0x0, 0x0);

				// because we want to write in the data register, we have to set the flag (0x80) (using 1=0x1 byte to send)
				byte flagByte = 0x80 | 0x1;
				int ret = SendReportId(0x5, flagByte, line, 0x0, 0x0, 0x0, 0x0, 0x0);
				if (ret == 0)
					return false;

			}

			// for testing, finally print all available special chars:
			this.WriteLine(1, "  SP0: " + (char)0x0 + "   SP1: " + (char)0x1);
			this.WriteLine(2, "  SP2: " + (char)0x2 + "   SP3: " + (char)0x3);
			this.WriteLine(3, "  SP4: " + (char)0x4 + "   SP5: " + (char)0x5);
			this.WriteLine(4, "  SP6: " + (char)0x6 + "   SP7: " + (char)0x7);
			return true;
		}

		/// <summary>
		/// Sends a specific ReportID to the IOWarrior Interface #1
		/// See the IOWarrior Datasheet for more Information
		/// </summary>
		/// <param name="reportId">The ReportID</param>
		/// <param name="d1">1. byte - this byte is usually used for flags, while all other bytes will be used for data bytes</param>
		/// <param name="d2">2. byte</param>
		/// <param name="d3">3. byte</param>
		/// <param name="d4">4. byte</param>
		/// <param name="d5">5. byte</param>
		/// <param name="d6">6. byte</param>
		/// <param name="d7">7. byte</param>
		/// <returns>Number of bytes written.</returns>
		private int SendReportId(byte reportId, byte d1, byte d2, byte d3, byte d4, byte d5, byte d6, byte d7)
		{
			if(iowHandle == 0)
				throw new Exception( "IOWarrior not initialized" );

			byte[] reportidBytes = { reportId, d1, d2, d3, d4, d5, d6, d7 };
			return IOWLib.IowKitWrite(iowHandle, 1, ref reportidBytes[0], reportidBytes.Length);
		}

		/// <summary>
		/// Genearets the byteValue for initializing the display
		/// </summary>
		/// <returns></returns>
		/// <remarks>
		///		if you don't understand this, have a look on the datasheet 
		///		for your Display or the HD44780 controller. (NOT IoWarrior!)
		///	</remarks>
		protected byte GenerateInitialFunctionSetValue()
		{
			// bitwise logical operators will be used to calculate the value
			// we have some hardcodes values here (use 8bit databus and 5x7 Dot Display)

			bool bUse8bit = true; // false = display databus interface should be in 4-bit mode
			bool b5x10Dots = false; // false = dots are 5x7
			bool b2Lines = this.nRows > 1 ? true : false; // 20x4 displays use 2 lines too, in this case, line 1+3 and 2+4 will usually be handled like one single line by the HD44780 controller

			byte bRetVal = 0x20; // 0x20 = binary 00100000; initial value to indicate that this byte will be used to set the "Function Set" Instruction
			if( bUse8bit )
				bRetVal = (byte)((int)bRetVal | 0x10); // 0x10 = binary 00010000; 
			if( b2Lines )
				bRetVal = (byte)((int)bRetVal | 0x08); // 0x08 = binary 00001000;
			if( b5x10Dots )
				bRetVal = (byte)((int)bRetVal | 0x04); // 0x04 = binary 00000100;

			return bRetVal;

		}

		/// <summary>
		/// Genearets the byteValue for display properties
		/// </summary>
		/// <returns></returns>
		/// <remarks>
		///		if you don't understand this, have a look on the datasheet 
		///		for your Display or the HD44780 controller. (NOT IoWarrior!)
		///	</remarks>
		protected byte GenerateInitialModeValue()
		{
			// bitwise logical operators will be used to calculate the value
			// bDisplayOn is hardcoded here

			bool bDisplayOn = true; // false = Display Off .. but i think, some displays will ignore this bit
			
			byte bRetVal = 0x8; // 0x8 = binary 00001000; initial value to indicate that this byte will be set the "Display On/Off Control" Instruction
			if( bDisplayOn )
				bRetVal = (byte)((int)bRetVal | 0x4); // 0x4 = binary 00000100;
			if( bShowCursor )
				bRetVal = (byte)((int)bRetVal | 0x2); // 0x2 = binary 00000010;
			if( bCursorBlinking )
				bRetVal = (byte)((int)bRetVal | 0x1); // 0x1 = binary 00000001;
			
			return bRetVal;

		}

		/// <summary>
		/// because the charset on the LCD does not match the ASCII codes, 
		/// we have to translate some characters here.
		/// have a look at your LCD Datasheet for your right charset.
		/// </summary>
		/// <param name="charValue">the character which should be translated</param>
		/// <returns>the correct byte-Value of the character</returns>
		protected byte GetLCDByteValue( char charValue )
		{
			if( charValue == '')
				return 0xe1;
			if( charValue == '')
				return 0xef;
			if( charValue == '')
				return 0xf5;
			if( charValue == '')
				return 0xe2;

			return (byte)charValue;
		}
	}
}
